<?php

namespace Drupal\devportal_api_entities\Entity;

use Drupal\Core\Entity\EntityChangedTrait;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\RevisionableContentEntityBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\devportal_api_reference\Traits\APIRefTrait;
use Drupal\devportal_api_entities\Traits\AutoLabelTrait;
use Drupal\devportal_api_entities\APIContactInterface;
use Drupal\devportal\Traits\URLRouteParametersTrait;;
use Drupal\devportal_api_entities\Traits\VendorExtensionTrait;
use Drupal\link\LinkItemInterface;
use Drupal\devportal_api_entities\Traits\APIVersionTagRefTrait;

/**
 * Defines the API Contact entity class.
 *
 * @ContentEntityType(
 *   id = "api_contact",
 *   label = @Translation("API Contact"),
 *   handlers = {
 *     "storage" = "Drupal\devportal_api_entities\APIContactStorage",
 *     "list_builder" = "Drupal\devportal_api_entities\APIContactListBuilder",
 *     "view_builder" = "Drupal\devportal\DevportalContentEntityViewBuilder",
 *     "views_data" = "Drupal\devportal_api_entities\APIContactViewsData",
 *     "form" = {
 *       "default" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "add" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "edit" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "delete" = "Drupal\devportal\Form\DevportalContentEntityDeleteForm",
 *     },
 *     "inline_form" = "Drupal\devportal_api_entities\Form\DevportalInlineForm",
 *     "route_provider" = {
 *       "html" = "Drupal\devportal_api_entities\APIContactHtmlRouteProvider",
 *     },
 *     "access" = "Drupal\devportal_api_entities\APIContactAccessControlHandler",
 *     "translation" = "Drupal\devportal_api_entities\APIContactTranslationHandler",
 *   },
 *   admin_permission = "administer api contacts",
 *   fieldable = TRUE,
 *   base_table = "api_contact",
 *   data_table = "api_contact_field_data",
 *   field_ui_base_route = "entity.api_contact_type.edit_form",
 *   entity_keys = {
 *     "id" = "id",
 *     "bundle" = "type",
 *     "uuid" = "uuid",
 *     "revision" = "vid",
 *     "langcode" = "langcode",
 *     "label" = "auto_label",
 *   },
 *   api_extra_info = {
 *     "auto_label" = {
 *       "auto_label" = "auto_label",
 *       "autogenerated_label" = "autogenerated_label",
 *     },
 *     "vendor_extension" = "extensions",
 *     "api_ref" = "api_ref",
 *     "api_version_tag" = "api_version_tag",
 *   },
 *   bundle_entity_type = "api_contact_type",
 *   bundle_label = @Translation("API Contact type"),
 *   revision_table = "api_contact_revision",
 *   revision_data_table = "api_contact_field_revision",
 *   show_revision_ui = TRUE,
 *   revision_metadata_keys = {
 *     "revision_created" = "revision_created",
 *     "revision_user" = "revision_user",
 *     "revision_log_message" = "revision_log",
 *   },
 *   links = {
 *     "canonical" = "/api_contact/{api_contact}",
 *     "add-page" = "/api_contact/add",
 *     "add-form" = "/api_contact/add/{api_contact_type}",
 *     "edit-form" = "/api_contact/{api_contact}/edit",
 *     "delete-form" = "/api_contact/{api_contact}/delete",
 *     "collection" = "/admin/content/api_contact",
 *     "version-history" = "/api_contact/{api_contact}/revisions",
 *     "revision" = "/api_contact/{api_contact}/revisions/{api_contact_revision}/view",
 *     "revision_revert" = "/api_contact/{api_contact}/revisions/{api_contact_revision}/revert",
 *     "revision_delete" = "/api_contact/{api_contact}/revisions/{api_contact_revision}/delete",
 *     "multiple_delete_confirm" = "/admin/content/api_contact/delete",
 *     "translation_revert" = "/api_contact/{api_contact}/revisions/{api_contact_revision}/revert/{langcode}",
 *   },
 *   translatable = TRUE,
 * )
 */
class APIContact extends RevisionableContentEntityBase implements APIContactInterface {

  use EntityChangedTrait;
  use VendorExtensionTrait;
  use AutoLabelTrait;
  use APIRefTrait;
  use APIVersionTagRefTrait;
  use URLRouteParametersTrait;

  /**
   * {@inheritdoc}
   */
  public function preSave(EntityStorageInterface $storage) {
    parent::preSave($storage);

    // Generate auto label.
    $this->autoLabelPreSave();
  }

  /**
   * {@inheritdoc}
   */
  public function preSaveRevision(EntityStorageInterface $storage, \stdClass $record) {
    parent::preSaveRevision($storage, $record);

    if (!$this->isNewRevision() && isset($this->original) && (!isset($record->revision_log) || $record->revision_log === '')) {
      // If we are updating an existing APIContact without adding a new
      // revision, we need to make sure $entity->revision_log is reset whenever
      // it is empty. Therefore, this code allows us to avoid clobbering an
      // existing log entry with an empty one.
      $record->revision_log = $this->original->revision_log->value;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function generateAutoLabel() {
    // Try to generate the auto label from these fields in this order (use the
    // name field's value if present; if not, use url field's value, etc.
    $fields = [
      'name' => 'value',
      'url' => 'uri',
      'mail' => 'value',
    ];
    foreach ($fields as $field_name => $column) {
      if (!$this->get($field_name)->isEmpty()) {
        $autolabel = $this->get($field_name)->getValue();
        return $autolabel[0][$column];
      }
    }

    // If there's no field value set, use the first vendor extension's name.
    $additional = $this->getEntityType()->get('additional');
    $extensions = $this->get($additional['api_extra_info']['vendor_extension']);
    if (!$extensions->isEmpty()) {
      $autolabel = $extensions->getValue();
      return $autolabel[0]['name'];
    }

    // If there's nothing on this entity that's usable as the auto label, have
    // a fallback.
    return t('PLACEHOLDER');
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    /** @var \Drupal\Core\Field\BaseFieldDefinition[] $fields */
    $fields = parent::baseFieldDefinitions($entity_type);

    // Add the auto label field.
    $fields += static::autoLabelBaseFieldDefinitions($entity_type);

    // Add the vendor extension fields.
    $fields += static::vendorExtensionBaseFieldDefinitions($entity_type);

    // Add the API Reference field.
    $fields += static::apiRefBaseFieldDefinitions($entity_type);
    $fields['api_ref']->setDescription(t('API Reference referenced from API Contact.'));

    // Add the API Version Tag field.
    $fields += static::apiVersionTagBaseFieldDefinitions($entity_type);
    $fields['api_version_tag']->setDescription(t('API Version Tag referenced from API Contact.'));

    $fields['id']->setDescription(t('The API Contact ID.'));

    $fields['uuid']->setDescription(t('The API Contact UUID.'));

    $fields['vid']->setDescription(t('The API Contact revision ID.'));

    $fields['langcode']->setDescription(t('The API Contact language code.'));

    $fields['name'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Name'))
      ->setRevisionable(TRUE)
      ->setTranslatable(TRUE)
      ->setSetting('max_length', 255)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'string',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['url'] = BaseFieldDefinition::create('link')
      ->setLabel(t('URL'))
      ->setRevisionable(TRUE)
      ->setTranslatable(TRUE)
      ->setSettings([
        'link_type' => LinkItemInterface::LINK_GENERIC,
        // @see \Drupal\link\Plugin\Field\FieldType\LinkItem::fieldSettingsForm()
        // and don't ask whys.
        'title' => DRUPAL_DISABLED,
      ])
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'link',
        'weight' => -5,
        'settings' => [
          'trim_length' => 80,
          'url_only' => FALSE,
          'rel' => FALSE,
          'target' => '_blank',
        ],
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'link_default',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['mail'] = BaseFieldDefinition::create('email')
      ->setLabel(t('Email'))
      ->setRevisionable(TRUE)
      ->setTranslatable(TRUE)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'email_mailto',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'email_default',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['changed'] = BaseFieldDefinition::create('changed')
      ->setLabel(t('Changed'))
      ->setDescription(t('The time that the API Contact was last edited.'))
      ->setTranslatable(TRUE)
      ->setRevisionable(TRUE);

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function label() {
    $label = parent::label();
    if (empty($label)) {
      $label = $this->generateAutoLabel();
    }
    return $label;
  }

}
