<?php

namespace Drupal\devportal_api_entities\Entity;

use Drupal\Core\Entity\EntityChangedTrait;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\RevisionableContentEntityBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\devportal_api_reference\Traits\APIRefTrait;
use Drupal\devportal_api_entities\Traits\APIParamRefTrait;
use Drupal\devportal_api_entities\Traits\AutoLabelTrait;
use Drupal\devportal_api_entities\APIEndpointInterface;
use Drupal\devportal\Traits\URLRouteParametersTrait;;
use Drupal\devportal_api_entities\Traits\VendorExtensionTrait;
use Drupal\devportal_api_entities\Traits\APIVersionTagRefTrait;

/**
 * Defines the API Endpoint entity class.
 *
 * @ContentEntityType(
 *   id = "api_endpoint",
 *   label = @Translation("API Endpoint"),
 *   handlers = {
 *     "storage" = "Drupal\devportal_api_entities\APIEndpointStorage",
 *     "list_builder" = "Drupal\devportal_api_entities\APIEndpointListBuilder",
 *     "view_builder" = "Drupal\devportal\DevportalContentEntityViewBuilder",
 *     "views_data" = "Drupal\devportal_api_entities\APIEndpointViewsData",
 *     "form" = {
 *       "default" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "add" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "edit" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "delete" = "Drupal\devportal\Form\DevportalContentEntityDeleteForm",
 *     },
 *     "inline_form" = "Drupal\devportal_api_entities\Form\DevportalInlineForm",
 *     "route_provider" = {
 *       "html" = "Drupal\devportal_api_entities\APIEndpointHtmlRouteProvider",
 *     },
 *     "access" = "Drupal\devportal_api_entities\APIEndpointAccessControlHandler",
 *     "translation" = "Drupal\devportal_api_entities\APIEndpointTranslationHandler",
 *   },
 *   admin_permission = "administer api endpoints",
 *   fieldable = TRUE,
 *   base_table = "api_endpoint",
 *   data_table = "api_endpoint_field_data",
 *   field_ui_base_route = "entity.api_endpoint_type.edit_form",
 *   entity_keys = {
 *     "id" = "id",
 *     "bundle" = "type",
 *     "uuid" = "uuid",
 *     "revision" = "vid",
 *     "langcode" = "langcode",
 *     "label" = "auto_label",
 *   },
 *   api_extra_info = {
 *     "auto_label" = {
 *       "auto_label" = "auto_label",
 *       "autogenerated_label" = "autogenerated_label",
 *     },
 *     "vendor_extension" = "extensions",
 *     "api_ref" = "api_ref",
 *     "api_param" = "api_param",
 *     "api_version_tag" = "api_version_tag",
 *   },
 *   bundle_entity_type = "api_endpoint_type",
 *   bundle_label = @Translation("API Endpoint type"),
 *   revision_table = "api_endpoint_revision",
 *   revision_data_table = "api_endpoint_field_revision",
 *   show_revision_ui = TRUE,
 *   revision_metadata_keys = {
 *     "revision_created" = "revision_created",
 *     "revision_user" = "revision_user",
 *     "revision_log_message" = "revision_log",
 *   },
 *   links = {
 *     "canonical" = "/api_endpoint/{api_endpoint}",
 *     "add-page" = "/api_endpoint/add",
 *     "add-form" = "/api_endpoint/add/{api_endpoint_type}",
 *     "edit-form" = "/api_endpoint/{api_endpoint}/edit",
 *     "delete-form" = "/api_endpoint/{api_endpoint}/delete",
 *     "collection" = "/admin/content/api_endpoint",
 *     "version-history" = "/api_endpoint/{api_endpoint}/revisions",
 *     "revision" = "/api_endpoint/{api_endpoint}/revisions/{api_endpoint_revision}/view",
 *     "revision_revert" = "/api_endpoint/{api_endpoint}/revisions/{api_endpoint_revision}/revert",
 *     "revision_delete" = "/api_endpoint/{api_endpoint}/revisions/{api_endpoint_revision}/delete",
 *     "multiple_delete_confirm" = "/admin/content/api_endpoint/delete",
 *     "translation_revert" = "/api_endpoint/{api_endpoint}/revisions/{api_endpoint_revision}/revert/{langcode}",
 *   },
 *   translatable = TRUE,
 * )
 */
class APIEndpoint extends RevisionableContentEntityBase implements APIEndpointInterface {

  use EntityChangedTrait;
  use VendorExtensionTrait;
  use AutoLabelTrait;
  use APIRefTrait;
  use APIParamRefTrait;
  use APIVersionTagRefTrait;
  use URLRouteParametersTrait;

  /**
   * {@inheritdoc}
   */
  public function getUri() {
    return $this->get('uri')->value;
  }

  /**
   * {@inheritdoc}
   */
  public function setUri($uri) {
    $this->set('uri', $uri);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function preSave(EntityStorageInterface $storage) {
    parent::preSave($storage);

    // Generate auto label.
    $this->autoLabelPreSave();
  }

  /**
   * {@inheritdoc}
   */
  public function preSaveRevision(EntityStorageInterface $storage, \stdClass $record) {
    parent::preSaveRevision($storage, $record);

    if (!$this->isNewRevision() && isset($this->original) && (!isset($record->revision_log) || $record->revision_log === '')) {
      // If we are updating an existing APIEndpoint without adding a new
      // revision, we need to make sure $entity->revision_log is reset whenever
      // it is empty. Therefore, this code allows us to avoid clobbering an
      // existing log entry with an empty one.
      $record->revision_log = $this->original->revision_log->value;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function generateAutoLabel() {
    // Get the ID of the referenced API Reference entity.
    $api_ref_id = $this->get('api_ref')->getValue()[0]['target_id'];
    // Find the API Documentation entity which refers to the same API Reference
    // entity.
    $query = \Drupal::entityQuery('api_doc')
      ->condition('api_ref', $api_ref_id);
    $api_doc_ids = $query->execute();
    /** @var \Drupal\devportal_api_entities\Entity\APIDoc $api_doc */
    $api_doc = APIDoc::load(reset($api_doc_ids));

    return $api_doc->getBasePath() . $this->get('uri')->value;
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    /** @var \Drupal\Core\Field\BaseFieldDefinition[] $fields */
    $fields = parent::baseFieldDefinitions($entity_type);

    // Add the auto label field.
    $fields += static::autoLabelBaseFieldDefinitions($entity_type);

    // Add the vendor extension fields.
    $fields += static::vendorExtensionBaseFieldDefinitions($entity_type);

    // Add the API Reference field.
    $fields += static::apiRefBaseFieldDefinitions($entity_type);
    $fields['api_ref']->setDescription(t('API Reference referenced from API Endpoint.'));

    // Add the API Parameter field.
    $fields += static::apiParamBaseFieldDefinitions($entity_type);
    $fields['api_param']->setDescription(t('API Parameter referenced from API Endpoint.'));

    // Add the API Version Tag field.
    $fields += static::apiVersionTagBaseFieldDefinitions($entity_type);
    $fields['api_version_tag']->setDescription(t('API Version Tag referenced from API Endpoint.'));

    $fields['id']->setDescription(t('The API Endpoint ID.'));

    $fields['uuid']->setDescription(t('The API Endpoint UUID.'));

    $fields['vid']->setDescription(t('The API Endpoint revision ID.'));

    $fields['langcode']->setDescription(t('The API Endpoint language code.'));

    $fields['uri'] = BaseFieldDefinition::create('string')
      ->setLabel(t('URI'))
      ->setRequired(TRUE)
      ->setRevisionable(TRUE)
      ->setSetting('max_length', 255)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'string',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['api_endpoint_set'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Endpoint Set'))
      ->setRequired(TRUE)
      ->setRevisionable(TRUE)
      ->setDescription(t('API Endpoint Set referenced from API Endpoint.'))
      ->setSettings([
        'target_type' => 'api_endpoint_set',
        'handler' => 'default',
        'handler_settings' => [
          'sort' => [
            'field' => '_none',
          ],
          'auto_create' => FALSE,
          'auto_create_bundle' => '',
        ],
      ])
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'entity_reference_label',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'inline_entity_form_complex',
        'weight' => 5,
        'settings' => [
          'form_mode' => 'default',
          'override_labels' => TRUE,
          // @FIXME Should these use $this->>t()?
          'label_singular' => 'endpoint set',
          'label_plural' => 'endpoint sets',
          'allow_new' => TRUE,
          'allow_existing' => TRUE,
          'match_operator' => 'CONTAINS',
        ],
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['changed'] = BaseFieldDefinition::create('changed')
      ->setLabel(t('Changed'))
      ->setDescription(t('The time that the API Endpoint was last edited.'))
      ->setTranslatable(TRUE)
      ->setRevisionable(TRUE);

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function label() {
    $label = parent::label();
    if (empty($label)) {
      $label = $this->generateAutoLabel();
    }
    return $label;
  }

}
