<?php

namespace Drupal\devportal_api_entities\Entity;

use Drupal\Core\Entity\EntityChangedTrait;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\RevisionableContentEntityBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\devportal_api_reference\Traits\APIRefTrait;
use Drupal\devportal_api_entities\Traits\APIExtDocRefTrait;
use Drupal\devportal_api_entities\Traits\APIParamRefTrait;
use Drupal\devportal_api_entities\Traits\AutoLabelTrait;
use Drupal\devportal_api_entities\APIMethodInterface;
use Drupal\devportal\Traits\URLRouteParametersTrait;;
use Drupal\devportal_api_entities\Traits\VendorExtensionTrait;
use Drupal\devportal_api_entities\Traits\ProducesTrait;
use Drupal\devportal_api_entities\Traits\ConsumesTrait;
use Drupal\devportal_api_entities\Traits\APIVersionTagRefTrait;

/**
 * Defines the API Method entity class.
 *
 * @ContentEntityType(
 *   id = "api_method",
 *   label = @Translation("API HTTP Method"),
 *   handlers = {
 *     "storage" = "Drupal\devportal_api_entities\APIMethodStorage",
 *     "list_builder" = "Drupal\devportal_api_entities\APIMethodListBuilder",
 *     "view_builder" = "Drupal\devportal\DevportalContentEntityViewBuilder",
 *     "views_data" = "Drupal\devportal_api_entities\APIMethodViewsData",
 *     "form" = {
 *       "default" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "add" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "edit" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "delete" = "Drupal\devportal\Form\DevportalContentEntityDeleteForm",
 *     },
 *     "inline_form" = "Drupal\devportal_api_entities\Form\DevportalInlineForm",
 *     "route_provider" = {
 *       "html" = "Drupal\devportal_api_entities\APIMethodHtmlRouteProvider",
 *     },
 *     "access" = "Drupal\devportal_api_entities\APIMethodAccessControlHandler",
 *     "translation" = "Drupal\devportal_api_entities\APIMethodTranslationHandler",
 *   },
 *   admin_permission = "administer api methods",
 *   fieldable = TRUE,
 *   base_table = "api_method",
 *   data_table = "api_method_field_data",
 *   field_ui_base_route = "entity.api_method_type.edit_form",
 *   entity_keys = {
 *     "id" = "id",
 *     "bundle" = "type",
 *     "uuid" = "uuid",
 *     "revision" = "vid",
 *     "langcode" = "langcode",
 *     "label" = "auto_label",
 *   },
 *   api_extra_info = {
 *     "auto_label" = {
 *       "auto_label" = "auto_label",
 *       "autogenerated_label" = "autogenerated_label",
 *     },
 *     "vendor_extension" = "extensions",
 *     "produces" = "produces",
 *     "consumes" = "consumes",
 *     "api_ext_doc" = "api_ext_doc",
 *     "api_ref" = "api_ref",
 *     "api_param" = "api_param",
 *     "api_version_tag" = "api_version_tag",
 *   },
 *   bundle_entity_type = "api_method_type",
 *   bundle_label = @Translation("API HTTP Method type"),
 *   revision_table = "api_method_revision",
 *   revision_data_table = "api_method_field_revision",
 *   show_revision_ui = TRUE,
 *   revision_metadata_keys = {
 *     "revision_user" = "revision_user",
 *     "revision_created" = "revision_created",
 *     "revision_log_message" = "revision_log",
 *   },
 *   links = {
 *     "canonical" = "/api_method/{api_method}",
 *     "add-page" = "/api_method/add",
 *     "add-form" = "/api_method/add/{api_method_type}",
 *     "edit-form" = "/api_method/{api_method}/edit",
 *     "delete-form" = "/api_method/{api_method}/delete",
 *     "collection" = "/admin/content/api_method",
 *     "version-history" = "/api_method/{api_method}/revisions",
 *     "revision" = "/api_method/{api_method}/revisions/{api_method_revision}/view",
 *     "revision_revert" = "/api_method/{api_method}/revisions/{api_method_revision}/revert",
 *     "revision_delete" = "/api_method/{api_method}/revisions/{api_method_revision}/delete",
 *     "multiple_delete_confirm" = "/admin/content/api_method/delete",
 *     "translation_revert" = "/api_method/{api_method}/revisions/{api_method_revision}/revert/{langcode}",
 *   },
 *   translatable = TRUE,
 * )
 */
class APIMethod extends RevisionableContentEntityBase implements APIMethodInterface {

  use EntityChangedTrait;
  use VendorExtensionTrait;
  use AutoLabelTrait;
  use ProducesTrait;
  use ConsumesTrait;
  use APIExtDocRefTrait;
  use APIRefTrait;
  use APIParamRefTrait;
  use APIVersionTagRefTrait;
  use URLRouteParametersTrait;

  /**
   * {@inheritdoc}
   */
  public function getHTTPMethod() {
    return $this->get('http_method')->value;
  }

  /**
   * {@inheritdoc}
   */
  public function setHTTPMethod($version) {
    $this->set('http_method', $version);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() {
    return $this->get('description')->value;
  }

  /**
   * {@inheritdoc}
   */
  public function setDescription($version) {
    $this->set('description', $version);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function preSave(EntityStorageInterface $storage) {
    parent::preSave($storage);

    // Generate auto label.
    $this->autoLabelPreSave();
  }

  /**
   * {@inheritdoc}
   */
  public function preSaveRevision(EntityStorageInterface $storage, \stdClass $record) {
    parent::preSaveRevision($storage, $record);

    if (!$this->isNewRevision() && isset($this->original) && (!isset($record->revision_log) || $record->revision_log === '')) {
      // If we are updating an existing APIMethod without adding a new revision,
      // we need to make sure $entity->revision_log is reset whenever it is
      // empty. Therefore, this code allows us to avoid clobbering an existing
      // log entry with an empty one.
      $record->revision_log = $this->original->revision_log->value;
    }
  }

  /**
   * {@inheritdoc}
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  public function generateAutoLabel() {
    $methods = $this->get('http_method')->getSetting('allowed_values');

    /** @var \Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItem $endpoint_first */
    $endpoint_first = $this->get('api_endpoint')->first();
    /** @var \Drupal\Core\Entity\Plugin\DataType\EntityReference $endpoint_first_entity */
    $endpoint_first_entity = $endpoint_first->get('entity');
    $endpoint = $endpoint_first_entity->getTarget()->getValue();

    return $methods[$this->get('http_method')->value] . ' ' . $endpoint->label();
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    /** @var \Drupal\Core\Field\BaseFieldDefinition[] $fields */
    $fields = parent::baseFieldDefinitions($entity_type);

    // Add the auto label field.
    $fields += static::autoLabelBaseFieldDefinitions($entity_type);

    // Add the vendor extension fields.
    $fields += static::vendorExtensionBaseFieldDefinitions($entity_type);

    // Add the produces fields.
    $fields += static::producesBaseFieldDefinitions($entity_type);

    // Add the consumes fields.
    $fields += static::consumesBaseFieldDefinitions($entity_type);

    // Add the API External Documentation field.
    $fields += static::apiExtDocBaseFieldDefinitions($entity_type);
    $fields['api_ext_doc']->setDescription(t('API External Documentation referenced from API Method.'));

    // Add the API Reference field.
    $fields += static::apiRefBaseFieldDefinitions($entity_type);
    $fields['api_ref']->setDescription(t('API Reference referenced from API Method.'));

    // Add the API Parameter field.
    $fields += static::apiParamBaseFieldDefinitions($entity_type);
    $fields['api_param']->setDescription(t('API Parameter referenced from API Method.'));

    // Add the API Version Tag field.
    $fields += static::apiVersionTagBaseFieldDefinitions($entity_type);
    $fields['api_version_tag']->setDescription(t('API Version Tag referenced from API Method.'));

    $fields['id']->setDescription(t('The API HTTP Method ID.'));

    $fields['uuid']->setDescription(t('The API HTTP Method UUID.'));

    $fields['vid']->setDescription(t('The API HTTP Method revision ID.'));

    $fields['langcode']->setDescription(t('The API HTTP Method language code.'));

    $fields['http_method'] = BaseFieldDefinition::create('list_string')
      ->setLabel(t('HTTP Method'))
      ->setRequired(TRUE)
      ->setRevisionable(TRUE)
      ->setSettings([
        'allowed_values' => [
          'get' => 'GET',
          'post' => 'POST',
          'head' => 'HEAD',
          'put' => 'PUT',
          'delete' => 'DELETE',
          'options' => 'OPTIONS',
          'patch' => 'PATCH',
        ]
      ])
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'string',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'options_select',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['description'] = BaseFieldDefinition::create('text_long')
      ->setLabel(t('Description'))
      ->setTranslatable(TRUE)
      ->setRevisionable(TRUE)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'text_default',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'text_textfield',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['summary'] = BaseFieldDefinition::create('string_long')
      ->setLabel(t('Summary'))
      ->setTranslatable(TRUE)
      ->setRevisionable(TRUE)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'text_default',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'text_textfield',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['api_endpoint'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Endpoint'))
      ->setRequired(TRUE)
      ->setRevisionable(TRUE)
      ->setDescription(t('API Endpoint referenced from API Method.'))
      ->setSettings([
        'target_type' => 'api_endpoint',
        'handler' => 'default',
        'handler_settings' => [
          'sort' => [
            'field' => '_none',
          ],
          'auto_create' => FALSE,
          'auto_create_bundle' => '',
        ],
      ])
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'entity_reference_label',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'inline_entity_form_complex',
        'weight' => 5,
        'settings' => [
          'form_mode' => 'default',
          'override_labels' => TRUE,
          // @FIXME Should these use $this->>t()?
          'label_singular' => 'endpoint',
          'label_plural' => 'endpoints',
          'allow_new' => TRUE,
          'allow_existing' => TRUE,
          'match_operator' => 'CONTAINS',
        ],
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['op_id'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Operation ID'))
      ->setRevisionable(TRUE)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'string',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['deprecated'] = BaseFieldDefinition::create('boolean')
      ->setLabel(t('Deprecated'))
      ->setRevisionable(TRUE)
      ->setSettings([
        'on_label' => t('Yes'),
        'off_label' => t('No'),
      ])
      ->setDisplayOptions('view', [
        'label' => 'inline',
        'type' => 'boolean',
        'settings' => [
          'format' => 'default',
        ],
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'boolean_checkbox',
        'settings' => [
          'display_label' => TRUE,
        ],
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['api_tag'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('API Tags'))
      ->setRevisionable(TRUE)
      ->setDescription(t('API Tags referenced from API Method.'))
      ->setCardinality(BaseFieldDefinition::CARDINALITY_UNLIMITED)
      ->setSettings([
        'target_type' => 'api_tag',
        'handler' => 'default',
        'handler_settings' => [
          'sort' => [
            'field' => '_none',
          ],
          'auto_create' => FALSE,
          'auto_create_bundle' => '',
        ],
      ])
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'entity_reference_label',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'inline_entity_form_complex',
        'weight' => 5,
        'settings' => [
          'form_mode' => 'default',
          'override_labels' => TRUE,
          // @FIXME Should these use $this->>t()?
          'label_singular' => 'API Tag',
          'label_plural' => 'API Tags',
          'allow_new' => TRUE,
          'allow_existing' => TRUE,
          'match_operator' => 'CONTAINS',
        ],
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['changed'] = BaseFieldDefinition::create('changed')
      ->setLabel(t('Changed'))
      ->setDescription(t('The time that the API HTTP Method was last edited.'))
      ->setTranslatable(TRUE)
      ->setRevisionable(TRUE);

    return $fields;
  }

  /**
   * {@inheritdoc}
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  public function label() {
    $label = parent::label();
    if (empty($label)) {
      $label = $this->generateAutoLabel();
    }
    return $label;
  }

}
