<?php

namespace Drupal\devportal_api_entities\Entity;

use Drupal\Core\Entity\EntityChangedTrait;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\RevisionableContentEntityBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\devportal_api_entities\APIResponseInterface;
use Drupal\devportal_api_reference\Traits\APIRefTrait;
use Drupal\devportal_api_entities\Traits\AutoLabelTrait;
use Drupal\devportal_api_entities\Traits\APIVersionTagRefTrait;
use Drupal\devportal\Traits\URLRouteParametersTrait;;
use Drupal\devportal_api_entities\Traits\VendorExtensionTrait;

/**
 * Defines the API Response entity class.
 *
 * @ContentEntityType(
 *   id = "api_response",
 *   label = @Translation("API Response"),
 *   handlers = {
 *     "storage" = "Drupal\devportal_api_entities\APIResponseStorage",
 *     "list_builder" = "Drupal\devportal_api_entities\APIResponseListBuilder",
 *     "view_builder" = "Drupal\devportal\DevportalContentEntityViewBuilder",
 *     "views_data" = "Drupal\devportal_api_entities\APIResponseViewsData",
 *     "form" = {
 *       "default" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "add" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "edit" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "delete" = "Drupal\devportal\Form\DevportalContentEntityDeleteForm",
 *     },
 *     "inline_form" = "Drupal\devportal_api_entities\Form\DevportalInlineForm",
 *     "route_provider" = {
 *       "html" = "Drupal\devportal_api_entities\APIResponseHtmlRouteProvider",
 *     },
 *     "access" = "Drupal\devportal_api_entities\APIResponseAccessControlHandler",
 *     "translation" = "Drupal\devportal_api_entities\APIResponseTranslationHandler",
 *   },
 *   admin_permission = "administer api responses",
 *   fieldable = TRUE,
 *   base_table = "api_response",
 *   data_table = "api_response_field_data",
 *   field_ui_base_route = "entity.api_response_type.edit_form",
 *   entity_keys = {
 *     "id" = "id",
 *     "bundle" = "type",
 *     "uuid" = "uuid",
 *     "revision" = "vid",
 *     "langcode" = "langcode",
 *     "label" = "auto_label",
 *   },
 *   api_extra_info = {
 *     "auto_label" = {
 *       "auto_label" = "auto_label",
 *       "autogenerated_label" = "autogenerated_label",
 *     },
 *     "vendor_extension" = "extensions",
 *     "api_ref" = "api_ref",
 *     "api_version_tag" = "api_version_tag",
 *   },
 *   bundle_entity_type = "api_response_type",
 *   bundle_label = @Translation("API Response type"),
 *   revision_table = "api_response_revision",
 *   revision_data_table = "api_response_field_revision",
 *   show_revision_ui = TRUE,
 *   revision_metadata_keys = {
 *     "revision_user" = "revision_user",
 *     "revision_created" = "revision_created",
 *     "revision_log_message" = "revision_log",
 *   },
 *   links = {
 *     "canonical" = "/api_response/{api_response}",
 *     "add-page" = "/api_response/add",
 *     "add-form" = "/api_response/add/{api_response_type}",
 *     "edit-form" = "/api_response/{api_response}/edit",
 *     "delete-form" = "/api_response/{api_response}/delete",
 *     "collection" = "/admin/content/api_response",
 *     "version-history" = "/api_response/{api_response}/revisions",
 *     "revision" = "/api_response/{api_response}/revisions/{api_response_revision}/view",
 *     "revision_revert" = "/api_response/{api_response}/revisions/{api_response_revision}/revert",
 *     "revision_delete" = "/api_response/{api_response}/revisions/{api_response_revision}/delete",
 *     "multiple_delete_confirm" = "/admin/content/api_response/delete",
 *     "translation_revert" = "/api_response/{api_response}/revisions/{api_response_revision}/revert/{langcode}",
 *   },
 *   translatable = TRUE,
 * )
 */
class APIResponse extends RevisionableContentEntityBase implements APIResponseInterface {

  use EntityChangedTrait;
  use VendorExtensionTrait;
  use AutoLabelTrait;
  use APIRefTrait;
  use APIVersionTagRefTrait;
  use URLRouteParametersTrait;

  /**
   * {@inheritdoc}
   */
  public function getCode() {
    return $this->get('code')->value;
  }

  /**
   * {@inheritdoc}
   */
  public function setCode($code) {
    $this->set('code', $code);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() {
    return $this->get('description')->value;
  }

  /**
   * {@inheritdoc}
   */
  public function setDescription($description) {
    $this->set('description', $description);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function preSave(EntityStorageInterface $storage) {
    parent::preSave($storage);

    // Generate auto label.
    $this->autoLabelPreSave();
  }

  /**
   * {@inheritdoc}
   */
  public function preSaveRevision(EntityStorageInterface $storage, \stdClass $record) {
    parent::preSaveRevision($storage, $record);

    if (!$this->isNewRevision() && isset($this->original) && (!isset($record->revision_log) || $record->revision_log === '')) {
      // If we are updating an existing APIResponse without adding a new
      // revision, we need to make sure $entity->revision_log is reset whenever
      // it is empty. Therefore, this code allows us to avoid clobbering an
      // existing log entry with an empty one.
      $record->revision_log = $this->original->revision_log->value;
    }
  }

  /**
   * {@inheritdoc}
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  public function generateAutoLabel() {
    if ($this->getDescription()) {
      return $this->getCode() . ' ' . $this->getDescription();
    }

    /** @var \Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItem $global_response_first */
    if ($global_response_first = $this->get('api_global_response')->first()) {
      /** @var \Drupal\Core\Entity\Plugin\DataType\EntityReference $global_response_first_entity */
      $global_response_first_entity = $global_response_first->get('entity');
      /** @var \Drupal\devportal_api_entities\Entity\APIGlobalResponse $global_response */
      $global_response = $global_response_first_entity->getTarget()->getValue();
      return $this->getCode() . ' ' . $global_response->getDescription();
    }

    return $this->getCode();
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    /** @var \Drupal\Core\Field\BaseFieldDefinition[] $fields */
    $fields = parent::baseFieldDefinitions($entity_type);

    // Add the auto label field.
    $fields += static::autoLabelBaseFieldDefinitions($entity_type);

    // Add the vendor extension fields.
    $fields += static::vendorExtensionBaseFieldDefinitions($entity_type);

    // Add the API Reference field.
    $fields += static::apiRefBaseFieldDefinitions($entity_type);
    $fields['api_ref']->setDescription(t('API Reference referenced from API Response.'));

    // Add the API Version Tag field.
    $fields += static::apiVersionTagBaseFieldDefinitions($entity_type);
    $fields['api_version_tag']->setDescription(t('API Version Tag referenced from API Response.'));

    $fields['id']->setDescription(t('The API Response ID.'));

    $fields['uuid']->setDescription(t('The API Response UUID.'));

    $fields['vid']->setDescription(t('The API Response revision ID.'));

    $fields['langcode']->setDescription(t('The API Response language code.'));

    $fields['code'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Code'))
      ->setRequired(TRUE)
      ->setRevisionable(TRUE)
      ->setSetting('unsigned', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'string',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['description'] = BaseFieldDefinition::create('text_long')
      ->setLabel(t('Description'))
      ->setTranslatable(TRUE)
      ->setRevisionable(TRUE)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'text_default',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'text_textfield',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['api_schema'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('API Schema'))
      ->setDescription(t('API Schema referenced from API Response.'))
      ->setRevisionable(TRUE)
      ->setSettings([
        'target_type' => 'api_schema',
        'handler' => 'default',
        'handler_settings' => [
          'sort' => [
            'field' => '_none',
          ],
          'auto_create' => FALSE,
          'auto_create_bundle' => '',
        ],
      ])
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'entity_reference_label',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'inline_entity_form_complex',
        'weight' => 5,
        'settings' => [
          'form_mode' => 'default',
          'override_labels' => TRUE,
          // @FIXME Should these use $this->>t()?
          'label_singular' => 'API Schema',
          'label_plural' => 'API Schemas',
          'allow_new' => TRUE,
          'allow_existing' => FALSE,
          'match_operator' => 'CONTAINS',
        ],
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['api_response_set'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Response Set'))
      ->setRequired(TRUE)
      ->setRevisionable(TRUE)
      ->setDescription(t('API Response Set referenced from API Response.'))
      ->setSettings([
        'target_type' => 'api_response_set',
        'handler' => 'default',
        'handler_settings' => [
          'sort' => [
            'field' => '_none',
          ],
          'auto_create' => FALSE,
          'auto_create_bundle' => '',
        ],
      ])
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'entity_reference_label',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'inline_entity_form_complex',
        'weight' => 5,
        'settings' => [
          'form_mode' => 'default',
          'override_labels' => TRUE,
          // @FIXME Should these use $this->>t()?
          'label_singular' => 'response set',
          'label_plural' => 'response sets',
          'allow_new' => TRUE,
          'allow_existing' => TRUE,
          'match_operator' => 'CONTAINS',
        ],
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['api_global_response'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Global Response'))
      ->setRevisionable(TRUE)
      ->setDescription(t('API Global Response referenced from API Response.'))
      ->setSettings([
        'target_type' => 'api_global_response',
        'handler' => 'default',
        'handler_settings' => [
          'sort' => [
            'field' => '_none',
          ],
          'auto_create' => FALSE,
          'auto_create_bundle' => '',
        ],
      ])
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'entity_reference_label',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'inline_entity_form_complex',
        'weight' => 5,
        'settings' => [
          'form_mode' => 'default',
          'override_labels' => TRUE,
          // @FIXME Should these use $this->>t()?
          'label_singular' => 'global response',
          'label_plural' => 'global responses',
          'allow_new' => TRUE,
          'allow_existing' => TRUE,
          'match_operator' => 'CONTAINS',
        ],
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['changed'] = BaseFieldDefinition::create('changed')
      ->setLabel(t('Changed'))
      ->setDescription(t('The time that the API Response was last edited.'))
      ->setTranslatable(TRUE)
      ->setRevisionable(TRUE);

    return $fields;
  }

  /**
   * {@inheritdoc}
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  public function label() {
    $label = parent::label();
    if (empty($label)) {
      $label = $this->generateAutoLabel();
    }
    return $label;
  }

}
