<?php

namespace Drupal\devportal_api_entities\Entity;

use Drupal\Core\Entity\EntityChangedTrait;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\RevisionableContentEntityBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\devportal_api_reference\Traits\APIRefTrait;
use Drupal\devportal_api_entities\Traits\AutoLabelTrait;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\devportal_api_entities\APISchemaInterface;
use Drupal\devportal_api_entities\Traits\APIExtDocRefTrait;
use Drupal\devportal_api_entities\Traits\APIVersionTagRefTrait;
use Drupal\devportal\Traits\URLRouteParametersTrait;;

/**
 * Defines the API Schema entity class.
 *
 * @ContentEntityType(
 *   id = "api_schema",
 *   label = @Translation("API Schema"),
 *   handlers = {
 *     "storage" = "Drupal\devportal_api_entities\APISchemaStorage",
 *     "list_builder" = "Drupal\devportal_api_entities\APISchemaListBuilder",
 *     "view_builder" = "Drupal\devportal\DevportalContentEntityViewBuilder",
 *     "views_data" = "Drupal\devportal_api_entities\APISchemaViewsData",
 *     "form" = {
 *       "default" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "add" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "edit" = "Drupal\devportal\Form\DevportalContentEntityForm",
 *       "delete" = "Drupal\devportal\Form\DevportalContentEntityDeleteForm",
 *     },
 *     "inline_form" = "Drupal\devportal_api_entities\Form\DevportalInlineForm",
 *     "route_provider" = {
 *       "html" = "Drupal\devportal_api_entities\APISchemaHtmlRouteProvider",
 *     },
 *     "access" = "Drupal\devportal_api_entities\APISchemaAccessControlHandler",
 *     "translation" = "Drupal\devportal_api_entities\APISchemaTranslationHandler",
 *   },
 *   admin_permission = "administer api schemas",
 *   fieldable = TRUE,
 *   base_table = "api_schema",
 *   data_table = "api_schema_field_data",
 *   field_ui_base_route = "entity.api_schema_type.edit_form",
 *   entity_keys = {
 *     "id" = "id",
 *     "bundle" = "type",
 *     "uuid" = "uuid",
 *     "revision" = "vid",
 *     "langcode" = "langcode",
 *     "label" = "auto_label",
 *   },
 *   api_extra_info = {
 *     "auto_label" = {
 *       "auto_label" = "auto_label",
 *       "autogenerated_label" = "autogenerated_label",
 *     },
 *     "auto_label_source" = "schema_id",
 *     "api_ext_doc" = "api_ext_doc",
 *     "api_ref" = "api_ref",
 *     "api_version_tag" = "api_version_tag",
 *   },
 *   bundle_entity_type = "api_schema_type",
 *   bundle_label = @Translation("API Schema type"),
 *   revision_table = "api_schema_revision",
 *   revision_data_table = "api_schema_field_revision",
 *   show_revision_ui = TRUE,
 *   revision_metadata_keys = {
 *     "revision_user" = "revision_user",
 *     "revision_created" = "revision_created",
 *     "revision_log_message" = "revision_log",
 *   },
 *   links = {
 *     "canonical" = "/api_schema/{api_schema}",
 *     "add-page" = "/api_schema/add",
 *     "add-form" = "/api_schema/add/{api_schema_type}",
 *     "edit-form" = "/api_schema/{api_schema}/edit",
 *     "delete-form" = "/api_schema/{api_schema}/delete",
 *     "collection" = "/admin/content/api_schema",
 *     "version-history" = "/api_schema/{api_schema}/revisions",
 *     "revision" = "/api_schema/{api_schema}/revisions/{api_schema_revision}/view",
 *     "revision_revert" = "/api_schema/{api_schema}/revisions/{api_schema_revision}/revert",
 *     "revision_delete" = "/api_schema/{api_schema}/revisions/{api_schema_revision}/delete",
 *     "multiple_delete_confirm" = "/admin/content/api_schema/delete",
 *     "translation_revert" = "/api_schema/{api_schema}/revisions/{api_schema_revision}/revert/{langcode}",
 *   },
 *   translatable = TRUE,
 *   constraints = {
 *     "APISchema" = {},
 *   },
 * )
 */
class APISchema extends RevisionableContentEntityBase implements APISchemaInterface {

  use EntityChangedTrait;
  use AutoLabelTrait;
  use APIExtDocRefTrait;
  use APIRefTrait;
  use APIVersionTagRefTrait;
  use URLRouteParametersTrait;

  /**
   * {@inheritdoc}
   */
  public function getSchemaID() {
    return $this->get('schema_id')->value;
  }

  /**
   * {@inheritdoc}
   */
  public function setSchemaID($schema_id) {
    $this->set('schema_id', $schema_id);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function preSave(EntityStorageInterface $storage) {
    parent::preSave($storage);

    // Generate auto label.
    $this->autoLabelPreSave();
  }

  /**
   * {@inheritdoc}
   */
  public function preSaveRevision(EntityStorageInterface $storage, \stdClass $record) {
    parent::preSaveRevision($storage, $record);

    if (!$this->isNewRevision() && isset($this->original) && (!isset($record->revision_log) || $record->revision_log === '')) {
      // If we are updating an existing APISchema without adding a new
      // revision, we need to make sure $entity->revision_log is reset whenever
      // it is empty. Therefore, this code allows us to avoid clobbering an
      // existing log entry with an empty one.
      $record->revision_log = $this->original->revision_log->value;
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    /** @var \Drupal\Core\Field\BaseFieldDefinition[] $fields */
    $fields = parent::baseFieldDefinitions($entity_type);

    // Add the auto label field.
    $fields += static::autoLabelBaseFieldDefinitions($entity_type);

    // Add the API External Documentation field.
    $fields += static::apiExtDocBaseFieldDefinitions($entity_type);
    $fields['api_ext_doc']->setDescription(t('API External Documentation referenced from API Schema.'));

    // Add the API Reference field.
    $fields += static::apiRefBaseFieldDefinitions($entity_type);
    $fields['api_ref']->setDescription(t('API Reference referenced from API Schema.'));

    // Add the API Version Tag field.
    $fields += static::apiVersionTagBaseFieldDefinitions($entity_type);
    $fields['api_version_tag']->setDescription(t('API Version Tag referenced from API Schema.'));

    $fields['id']->setDescription(t('The API Schema ID.'));

    $fields['uuid']->setDescription(t('The API Schema UUID.'));

    $fields['vid']->setDescription(t('The API Schema revision ID.'));

    $fields['langcode']->setDescription(t('The API Schema language code.'));

    $fields['schema_id'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Schema ID'))
      ->setRequired(TRUE)
      ->setTranslatable(TRUE)
      ->setRevisionable(TRUE)
      ->setSetting('max_length', 255)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'string',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['api_global_schema'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Global Schema ref.'))
      ->setDescription(t('API Global Schema referenced from API Schema.'))
      ->setRevisionable(TRUE)
      ->setSettings([
        'target_type' => 'api_global_schema',
        'handler' => 'default',
        'handler_settings' => [
          'sort' => [
            'field' => '_none',
          ],
          'auto_create' => FALSE,
          'auto_create_bundle' => '',
        ],
      ])
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'entity_reference_label',
        'weight' => -5,
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', [
        'type' => 'inline_entity_form_complex',
        'weight' => 5,
        'settings' => [
          'form_mode' => 'default',
          'override_labels' => TRUE,
          // @FIXME Should these use $this->>t()?
          'label_singular' => 'API Global Schema',
          'label_plural' => 'API Global Schemas',
          'allow_new' => TRUE,
          'allow_existing' => TRUE,
          'match_operator' => 'CONTAINS',
        ],
      ])
      ->setDisplayConfigurable('form', TRUE);

    // @TODO Serialized JSON blob.
    $fields['inline_schema_def'] = BaseFieldDefinition::create('string_long')
      ->setLabel(t('Inline schema def.'))
      ->setRevisionable(TRUE)
      ->setDescription(t('Inline definition of the API Schema.'))
      // @FIXME Does it need to be displayed, eg. via a special field formatter?
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'text_default',
        'weight' => 0,
      ])
      ->setDisplayConfigurable('view', TRUE)
      // @FIXME: Does it need any form widget at all? If not, set to read-only.
      ->setDisplayOptions('form', [
        'type' => 'text_textfield',
        'weight' => 0,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['changed'] = BaseFieldDefinition::create('changed')
      ->setLabel(t('Changed'))
      ->setDescription(t('The time that the API Schema was last edited.'))
      ->setTranslatable(TRUE)
      ->setRevisionable(TRUE);

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function label() {
    $label = parent::label();
    if (empty($label)) {
      $label = $this->generateAutoLabel();
    }
    return $label;
  }

}
